import pytest
import numpy as np
import pandas as pd
from collections import OrderedDict
from nglui.statebuilder import *


@pytest.fixture
def image_layer(img_path):
    img_layer = ImageLayerConfig(
        name="img",
        source=img_path,
    )
    return img_layer


@pytest.fixture
def seg_layer_basic(seg_path_precomputed):
    seg_layer = SegmentationLayerConfig(name="seg", source=seg_path_precomputed)
    return seg_layer


@pytest.fixture
def anno_layer_basic():
    anno_layer = AnnotationLayerConfig(name="anno")
    return anno_layer


def test_basic(image_layer, seg_layer_basic, anno_layer_basic):
    sb = StateBuilder([image_layer, seg_layer_basic, anno_layer_basic])
    state_url = sb.render_state()
    assert state_url[:4] == "http"

    state = sb.render_state(return_as="html")
    # padding for link html plus text
    assert state.data == f'<a href="{state_url}" target="_blank">Neuroglancer Link</a>'

    state = sb.render_state(return_as="dict")
    assert isinstance(state, OrderedDict)

    state = sb.render_state(return_as="json")
    assert type(state) is str

    state = sb.render_state(return_as="viewer")
    assert len(state.layer_names) == 3


def test_segmentation_layer(soma_df, seg_path_precomputed):
    seg_layer = SegmentationLayerConfig(
        name="seg",
        source=seg_path_precomputed,
        fixed_ids=[1000],
        selected_ids_column="pt_root_id",
    )

    sb = StateBuilder(layers=[seg_layer])
    state = sb.render_state(soma_df, return_as="dict")
    print(state["layers"])
    assert "648518346349538466" in state["layers"][0]["segments"]
    assert "1000" in state["layers"][0]["segments"]


def test_segmentation_layer_color(soma_df, seg_path_precomputed):
    soma_df["color"] = ["#fdd4c2", "#fca082", "#fb694a", "#e32f27", "#b11218"]
    seg_layer = SegmentationLayerConfig(
        name="seg",
        source=seg_path_precomputed,
        selected_ids_column="pt_root_id",
        color_column="color",
    )
    sb = StateBuilder(layers=[seg_layer])
    state = sb.render_state(soma_df, return_as="dict")
    print(state["layers"])
    assert state["layers"][0]["segmentColors"]["648518346349538715"] == "#e32f27"


def test_segmentation_layer_options(soma_df, seg_path_precomputed):
    segmentation_view_options = {"alpha_selected": 0.6, "alpha_3d": 0.2}
    seg_layer = SegmentationLayerConfig(
        seg_path_precomputed, view_kws=segmentation_view_options
    )
    sb = StateBuilder([seg_layer])
    state = sb.render_state(return_as="dict")
    assert state["layers"][0]["selectedAlpha"] == 0.6


def test_annotations(pre_syn_df):
    points = PointMapper(point_column="pre_pt_position")
    points_2 = PointMapper(point_column="ctr_pt_position")
    anno_layer = AnnotationLayerConfig(name="annos", mapping_rules=[points, points_2])

    sb = StateBuilder([anno_layer])
    state = sb.render_state(pre_syn_df, return_as="dict")
    assert len(state["layers"][0]["annotations"]) == 10


def test_array_annotations():
    data = np.array([[1, 2, 3], [3, 4, 5], [6, 5, 3], [1, 2, 1]])
    anno_layer = AnnotationLayerConfig(
        name="annos", array_data=True, mapping_rules=PointMapper()
    )
    sb = StateBuilder([anno_layer])
    state = sb.render_state(data, return_as="dict")
    assert len(state["layers"][0]["annotations"]) == 4


def test_annotations_line(pre_syn_df):
    lines = LineMapper(
        point_column_a="pre_pt_position", point_column_b="post_pt_position"
    )
    anno_layer = AnnotationLayerConfig(name="annos", mapping_rules=lines)

    sb = StateBuilder([anno_layer])
    state = sb.render_state(pre_syn_df, return_as="dict")
    assert len(state["layers"][0]["annotations"]) == 5


def test_annotations_sphere(soma_df):
    soma_df["radius"] = 5000
    spheres = SphereMapper(center_column="pt_position", radius_column="radius")
    anno_layer = AnnotationLayerConfig(name="annos", mapping_rules=spheres)
    sb = StateBuilder([anno_layer])
    state = sb.render_state(soma_df, return_as="dict")
    assert len(state["layers"][0]["annotations"]) == 5


def test_annotations_description(soma_df):
    soma_df["desc"] = ["a", "b", "c", "d", "e"]
    points = PointMapper("pt_position", description_column="desc")
    anno_layer = AnnotationLayerConfig(name="annos", mapping_rules=points)
    sb = StateBuilder([anno_layer])
    state = sb.render_state(soma_df, return_as="dict")
    assert state["layers"][0]["annotations"][2]["description"] == "c"


def test_annotations_linked(soma_df, soma_df_Int64, seg_path_precomputed):
    seg_layer = SegmentationLayerConfig(seg_path_precomputed, name="seg")

    points = PointMapper("pt_position", linked_segmentation_column="pt_root_id")
    anno_layer = AnnotationLayerConfig(
        mapping_rules=points, linked_segmentation_layer="seg"
    )

    sb = StateBuilder([seg_layer, anno_layer])
    state = sb.render_state(soma_df, return_as="dict")
    assert "648518346349538715" in state["layers"][1]["annotations"][3]["segments"]

    state = sb.render_state(soma_df_Int64, return_as="dict")
    assert len(state["layers"][1]["annotations"][0]["segments"]) == 0


def test_annotation_tags(soma_df):
    tag_list = ["i", "e"]

    points = PointMapper("pt_position", tag_column="cell_type")
    anno_layer = AnnotationLayerConfig(mapping_rules=points, tags=tag_list)

    sb = StateBuilder([anno_layer])
    state = sb.render_state(soma_df, return_as="dict")
    assert '2' in state["layers"][0]["annotations"][1]["tagIds"]


def test_annotation_groups(pre_syn_df):
    df = pre_syn_df.copy()
    df["group"] = [1, 1, np.nan, 2.0, 2.0]
    points = PointMapper("ctr_pt_position", group_column="group")
    anno_layer = AnnotationLayerConfig(mapping_rules=[points])
    sb = StateBuilder([anno_layer])
    state = sb.render_state(df, return_as="dict")
    assert len(state["layers"][0]["annotations"]) == 7


def test_view_options(soma_df, image_layer):
    view_options = {
        "layout": "4panel",
        "show_slices": True,
        "zoom_3d": 500,
        "position": [71832, 54120, 1089],
    }

    sb = StateBuilder([image_layer], view_kws=view_options)
    state = sb.render_state(return_as="dict")
    assert state["navigation"]["pose"]["position"]["voxelCoordinates"] == [
        71832,
        54120,
        1089,
    ]

    points = PointMapper("pt_position", set_position=True)
    anno_layer = AnnotationLayerConfig(mapping_rules=points)

    sb = StateBuilder([image_layer, anno_layer])
    state = sb.render_state(soma_df, return_as="dict")
    assert state["navigation"]["pose"]["position"]["voxelCoordinates"] == list(
        soma_df["pt_position"].loc[0]
    )

def test_chained(pre_syn_df, post_syn_df, image_layer):
    # First state builder

    postsyn_mapper = LineMapper(
        point_column_a="pre_pt_position", point_column_b="ctr_pt_position"
    )
    postsyn_annos = AnnotationLayerConfig(
        "post", color="#00CCCC", mapping_rules=postsyn_mapper
    )

    postsyn_sb = StateBuilder(layers=[image_layer, postsyn_annos])

    # Second state builder
    presyn_mapper = LineMapper(
        point_column_a="ctr_pt_position", point_column_b="post_pt_position"
    )
    presyn_annos = AnnotationLayerConfig(
        "pre", color="#CC1111", mapping_rules=presyn_mapper
    )

    presyn_sb = StateBuilder(layers=[presyn_annos])

    # Chained state builder
    chained_sb = ChainedStateBuilder([postsyn_sb, presyn_sb])
    state = chained_sb.render_state([post_syn_df, pre_syn_df], return_as="dict")
    assert len(state["layers"]) == 3

def test_mapping_sets(pre_syn_df, post_syn_df, image_layer):
    postsyn_mapper = LineMapper(
        point_column_a="pre_pt_position", point_column_b="ctr_pt_position", mapping_set='post',
    )
    postsyn_annos = AnnotationLayerConfig(
        "post", color="#00CCCC", mapping_rules=postsyn_mapper
    )

    presyn_mapper = LineMapper(
        point_column_a="ctr_pt_position", point_column_b="post_pt_position", mapping_set='pre'
    )
    presyn_annos = AnnotationLayerConfig(
        "pre", color="#CC1111", mapping_rules=presyn_mapper
    )

    sb = StateBuilder([image_layer, postsyn_annos, presyn_annos])
    state = sb.render_state({'pre': pre_syn_df, 'post': post_syn_df}, return_as='dict')
    assert len(state['layers'][1]['annotations']) > 0
    assert len(state['layers'][2]['annotations']) > 0


def test_split_points(split_point_df, seg_path_graphene):
    split_mapper = SplitPointMapper(
        id_column="seg_id", point_column="pts", team_column="team", focus=True
    )
    seg = SegmentationLayerConfig(seg_path_graphene, split_point_map=split_mapper)
    sb = StateBuilder([seg])
    state = sb.render_state(split_point_df, return_as="dict")
    seg_layer = state["layers"][0]
    assert len(seg_layer["graphOperationMarker"][1]["annotations"]) == 2


def test_timestamp(seg_path_graphene):
    ts = 12345
    seg = SegmentationLayerConfig(seg_path_graphene, timestamp=ts)
    sb = StateBuilder([seg])
    state = sb.render_state(return_as="dict")
    seg_layer = state["layers"][0]
    assert int(seg_layer["timestamp"]) == 12345



from . import annotation, utils
from numpy import issubdtype, integer, uint64, vstack
from collections import OrderedDict
import copy
import re

SEGMENTATION_LAYER_TYPES = ["segmentation", "segmentation_with_graph"]


def _EasyViewerFactory(compatibility_mode=False):
    if compatibility_mode:
        import nglui.nglite_compatibility as neuroglancer
        from nglui.nglite_compatibility import annotation_compatibility
    else:
        import nglui.nglite as neuroglancer

    class EasyViewerSub(neuroglancer.Viewer):
        """
        Extends the neuroglancer Viewer object to make simple operations simple.
        """

        def __init__(self):
            super(neuroglancer.Viewer, self).__init__()
            if compatibility_mode:
                self.state = neuroglancer.viewer_state.ViewerState()

        def __repr__(self):
            return self.as_url()

        def _repr_html_(self):
            return '<a href="%s" target="_blank">Viewer</a>' % self.as_url()

        def load_url(self, url):
            """Load neuroglancer compatible url and updates viewer state

            Attributes:
            url (str): neuroglancer url

            """
            state = neuroglancer.parse_url(url)
            self.set_state(state)

        @staticmethod
        def _smart_add_segmentation_layer(s, layer_name, source, **kwargs):
            if re.search(r"^graphene://", source) is not None:
                s.layers[layer_name] = neuroglancer.ChunkedgraphSegmentationLayer(
                    source=source, **kwargs
                )
            elif re.search(r"^precomputed://", source) is not None:
                s.layers[layer_name] = neuroglancer.SegmentationLayer(
                    source=source, **kwargs
                )

        def add_layers(
            self,
            image_layers={},
            segmentation_layers={},
            annotation_layers={},
            resolution=None,
        ):
            with self.txn() as s:
                for ln, kws in image_layers.items():
                    s.layers[ln] = neuroglancer.ImageLayer(**kws)
                for ln, kws in segmentation_layers.items():
                    self._smart_add_segmentation_layer(s, **kws)
                for ln, kws in annotation_layers.items():
                    s.layers[ln] = neuroglancer.AnnotationLayer(**kws)
                if resolution is not None:
                    s.voxel_size = resolution
            pass

        def add_segmentation_layer(self, layer_name, source, **kwargs):
            """Add segmentation layer to viewer instance.

            Attributes:
                layer_name (str): name of layer to be displayed in neuroglancer ui.
                source (str): source of neuroglancer segment layer
            """
            with self.txn() as s:
                self._smart_add_segmentation_layer(
                    s, layer_name=layer_name, source=source, **kwargs
                )

        def add_image_layer(self, layer_name, source, **kwargs):
            """Add segmentation layer to viewer instance.

            Attributes:
                layer_name (str): name of layer to be displayed in neuroglancer ui.
                source (str): source of neuroglancer image layer
            """
            with self.txn() as s:
                s.layers[layer_name] = neuroglancer.ImageLayer(source=source, **kwargs)

        def set_resolution(self, resolution):
            if compatibility_mode:
                with self.txn() as s:
                    dims = neuroglancer.CoordinateSpace(
                        names=["x", "y", "z"],
                        scales=resolution,
                        units="nm",
                    )
                    s.dimensions = dims
            else:
                with self.txn() as s:
                    s.voxel_size = resolution

        def add_contrast_shader(self, layer_name, black=0.0, white=1.0):
            shader_text = f"#uicontrol float black slider(min=0, max=1, default={black})\n#uicontrol float white slider(min=0, max=1, default={white})\nfloat rescale(float value) {{\n  return (value - black) / (white - black);\n}}\nvoid main() {{\n  float val = toNormalized(getDataValue());\n  if (val < black) {{\n    emitRGB(vec3(0,0,0));\n  }} else if (val > white) {{\n    emitRGB(vec3(1.0, 1.0, 1.0));\n  }} else {{\n    emitGrayscale(rescale(val));\n  }}\n}}\n"
            self._update_layer_shader(layer_name, shader_text)

        def _update_layer_shader(self, layer_name, shader_text):
            with self.txn() as s:
                s.layers[layer_name]._json_data["shader"] = shader_text

        def set_state_server(self, state_server):
            with self.txn() as s:
                s._json_data["jsonStateServer"] = state_server

        def add_annotation_layer(
            self,
            layer_name=None,
            color=None,
            linked_segmentation_layer=None,
            filter_by_segmentation=False,
            brackets_show_segmentation=True,
            selection_shows_segmentation=True,
            tags=None,
        ):
            """Add annotation layer to the viewer instance.

            Attributes:
                layer_name (str): name of layer to be created
            """
            if layer_name is None:
                layer_name = "annos"

            if layer_name in [l.name for l in self.state.layers]:
                return

            if linked_segmentation_layer is None:
                filter_by_segmentation = None

            with self.txn() as s:
                if compatibility_mode:
                    new_layer = neuroglancer.LocalAnnotationLayer(
                        dimensions=self.state.dimensions,
                        linked_segmentation_layer=linked_segmentation_layer,
                    )
                else:
                    new_layer = neuroglancer.AnnotationLayer(
                        linked_segmentation_layer=linked_segmentation_layer,
                        filter_by_segmentation=filter_by_segmentation,
                        brackets_show_segmentation=brackets_show_segmentation,
                        selection_shows_segmentation=selection_shows_segmentation,
                    )
                s.layers.append(name=layer_name, layer=new_layer)
                if color is not None:
                    s.layers[layer_name].annotationColor = utils.parse_color(color)

            if tags is not None and not compatibility_mode:
                self.add_annotation_tags(layer_name=layer_name, tags=tags)

        def set_annotation_layer_color(self, layer_name, color):
            """Set the color for the annotation layer"""
            if layer_name in [l.name for l in self.state.layers]:
                with self.txn() as s:
                    s.layers[layer_name].annotationColor = utils.parse_color(color)
            else:
                pass

        def clear_annotation_layers(self, layer_names):
            with self.txn() as s:
                for ln in layer_names:
                    s.layers[ln].annotations._data = []

        def set_annotation_one_shot(self, ln_anno_dict):
            """
            ln_anno_dict is a layer_name to annotation list dict.
            """
            with self.txn() as s:
                for ln, annos in ln_anno_dict.items():
                    s.layers[ln].annotations._data = annos

        def add_annotations(self, layer_name, annotations):
            """Add annotations to a viewer instance, the type is specified.
            If layer name does not exist, add the layer

            Attributes:
                layer_name (str): name of layer to be displayed in neuroglancer ui.
                layer_type (str): can be: 'points, ellipse or line' only
            """
            if compatibility_mode:
                annotations = [annotation_compatibility.convert_annotation(anno) for anno in annotations]
            with self.txn() as s:
                s.layers[layer_name].annotations.extend(annotations)

        def add_multilayer_annotations(self, layer_anno_dict):
            """
            layer_anno_dict is a layer_name to annotation list dict.
            """
            with self.txn() as s:
                for ln, annos in layer_anno_dict.items():
                    if annos is None:
                        continue 
                    if compatibility_mode:
                        annos = [annotation_compatibility.convert_annotation(anno) for anno in annos]
                    s.layers[ln].annotations.extend(annos)

        def remove_annotations(self, layer_name, anno_ids):
            if isinstance(anno_ids, str):
                anno_ids = [anno_ids]
            try:
                with self.txn() as s:
                    el = len(s.layers[layer_name].annotations)
                    for anno in reversed(s.layers[layer_name].annotations):
                        el -= 1
                        if anno.id in anno_ids:
                            anno_ids.remove(anno.id)
                            s.layers[layer_name].annotations.pop(el)
                            if len(anno_ids) == 0:
                                break
            except:
                self.update_message("Could not remove annotation")

        def add_annotation_tags(self, layer_name, tags):
            """
            Add a list of tags to an annotation layer
            """
            if layer_name not in self.layer_names:
                raise ValueError("Layer is not an annotation layer")
            tag_list = [
                OrderedDict({"id": tag_id + 1, "label": label})
                for tag_id, label in enumerate(tags)
            ]
            with self.txn() as s:
                s.layers[layer_name]._json_data["annotationTags"] = tag_list

        def update_description(self, layer_id_dict, new_description):
            layer_id_dict = copy.deepcopy(layer_id_dict)
            with self.txn() as s:
                try:
                    for layer_name, id_list in layer_id_dict.items():
                        for anno in s.layers[layer_name].annotations:
                            if anno.id in id_list:
                                if anno.description is None:
                                    anno.description = new_description
                                else:
                                    anno.description = "{}\n{}".format(
                                        anno.description, new_description
                                    )
                                id_list.remove(anno.id)
                                if len(id_list) == 0:
                                    break
                except Exception as e:
                    print(e)
                    self.update_message("Could not update descriptions!")

        @property
        def url(self):
            return self.get_viewer_url()

        def as_url(self, prefix=None, as_html=False, link_text="Neuroglancer link"):
            if prefix is None:
                prefix = utils.default_neuroglancer_base
            ngl_url = neuroglancer.to_url(self.state, prefix=prefix)
            if as_html:
                return '<a href="{}" target="_blank">{}</a>'.format(ngl_url, link_text)
            else:
                return ngl_url

        def update_message(self, message):
            with self.config_state.txn() as s:
                if message is not None:
                    s.status_messages["status"] = message

        def set_selected_layer(self, layer_name, tool=None):
            if layer_name in self.layer_names:
                with self.txn() as s:
                    s._json_data["selectedLayer"] = OrderedDict(
                        layer=layer_name, visible=True
                    )
                    if tool is not None:
                        s.layers[layer_name]._json_data["tool"] = tool

        def get_selected_layer(self):
            state_json = self.state.to_json()
            try:
                selected_layer = state_json["selectedLayer"]["layer"]
            except:
                selected_layer = None
            return selected_layer

        def get_annotation(self, layer_name, aid):
            if self.state.layers[layer_name].type == "annotation":
                for anno in self.state.layers[layer_name].annotations:
                    if anno.id == aid:
                        return anno
                else:
                    return None
            else:
                return None

        def get_selected_annotation_id(self):
            layer_name = self.get_selected_layer()
            try:
                aid_data = self.state.layers[layer_name]._json_data[
                    "selectedAnnotation"
                ]
                if isinstance(aid_data, OrderedDict):
                    aid = aid_data["id"]
                else:
                    aid = aid_data
            except:
                aid = None
            return aid

        def select_annotation(self, layer_name, aid):
            if layer_name in self.layer_names:
                with self.txn() as s:
                    s.layers[layer_name]._json_data["selectedAnnotation"] = aid
            self.set_selected_layer(layer_name)

        @property
        def layer_names(self):
            return [l.name for l in self.state.layers]

        def selected_objects(self, segmentation_layer):
            return list(self.state.layers[segmentation_layer].segments)

        def add_selected_objects(self, segmentation_layer, oids, colors=None):
            if issubdtype(type(oids), integer):
                oids = [oids]

            with self.txn() as s:
                for oid in oids:
                    s.layers[segmentation_layer].segments.add(uint64(oid))
                if s.layers[segmentation_layer].type == "segmentation_with_graph":
                    s.layers[segmentation_layer].segmentQuery = ", ".join(
                        [str(x) for x in oids]
                    )
            if colors is not None:
                if isinstance(colors, dict):
                    self.assign_colors(segmentation_layer, colors)
                elif len(colors) == len(oids):
                    seg_colors = {str(oid): clr for oid, clr in zip(oids, colors)}
                    self.assign_colors(segmentation_layer, seg_colors)

        def get_mouse_coordinates(self, s):
            pos = s.mouse_voxel_coordinates
            return pos

        def set_view_options(
            self,
            show_slices=None,
            layout=None,
            show_axis_lines=None,
            show_scale_bar=None,
            orthographic=None,
            position=None,
            zoom_image=None,
            zoom_3d=None,
            background_color=None,
        ):
            """Set options relating to the neuroglancer view state. Only changes the values of the parameters provided.

            Parameters
            ----------
            show_slices : bool, optional
                Show slice cutout in the 3d view, by default None
            layout : str, optional
                Change the layout type ('xy', 'yz', 'zx', '3d', 'xy-3d', 'yz-3d', 'zx-3d', or '4panel'), by default None
            show_axis_lines : bool, optional
                Show the red/blue/green lines indicating the axis directions, by default None
            show_scale_bar : bool, optional
                Controls showing of scale bar, by default None
            orthographic : bool, optional
                Controls whether the 3d perspective view is orthographic or not, by default None
            position : list, optional
                Sets the location of center point of the view in Neuroglancer coordinates, by default None
            zoom_image : int, optional
                Sets the zoom factor for the imagery, by default None
            zoom_3d : int, optional
                Sets the zoom factor for the 3d view, by default None
            background_color : list or str, optional
                hex, rgb, or named color for the background of the 3d viewer, by default None
            """
            with self.txn() as s:
                if show_slices is not None:
                    s.showSlices = show_slices
                if layout is not None:
                    s.layout.type = layout
                if show_axis_lines is not None:
                    s.show_axis_lines = show_axis_lines
                if show_scale_bar is not None:
                    s.show_scale_bar = show_scale_bar
                if orthographic is not None:
                    s.layout.orthographic_projection = orthographic
                if position is not None:
                    if compatibility_mode:
                        s.position = position
                    else:
                        s.position.voxelCoordinates = position
                if zoom_image is not None:
                    if compatibility_mode:
                        s.cross_section_scale = zoom_image
                    else:
                        s.navigation.zoomFactor = zoom_image
                if zoom_3d is not None:
                    if compatibility_mode:
                        s.projectionScale = zoom_3d
                    else:
                        s.perspectiveZoom = zoom_3d
                if background_color is not None:
                    s.perspectiveViewBackgroundColor = utils.parse_color(background_color)

        def set_segmentation_view_options(
            self,
            layer_name,
            alpha_selected=None,
            alpha_3d=None,
            alpha_unselected=None,
        ):
            if self.state.layers[layer_name].type not in SEGMENTATION_LAYER_TYPES:
                return
            with self.txn() as s:
                l = s.layers[layer_name]
                if alpha_selected is not None:
                    l.selectedAlpha = alpha_selected
                if alpha_3d is not None:
                    l.objectAlpha = alpha_3d
                if alpha_unselected is not None:
                    l.notSelectedAlpha = alpha_unselected

        def set_timestamp(
            self,
            layer_name,
            timestamp=None,
        ):
            """Set timestamp of a segmentation layer

            Parameters
            ----------
            layer_name : str
                Name of a segmentation layer
            timestamp : float, optional
                Timestamp in unix epoch time (e.g. `time.time.now()` in python), by default None
            """
            if self.state.layers[layer_name].type != "segmentation_with_graph":
                return
            with self.txn() as s:
                l = s.layers[layer_name]
                if timestamp is not None:
                    l.timestamp = int(timestamp)
                else:
                    l.timestamp = None

        def assign_colors(self, layer_name, seg_colors):
            """Assign colors to root ids in a segmentation layer

            Parameters
            ----------
            layer_name : str,
                Segmentation layer name
            seg_colors : dict
                dict with root ids as keys and colors as values.
            """
            with self.txn() as s:
                if seg_colors is not None:
                    seg_colors = {
                        str(oid): utils.parse_color(k)
                        for oid, k in seg_colors.items()
                        if k is not None
                    }
                    s.layers[layer_name]._json_data["segmentColors"] = seg_colors

        def set_multicut_points(
            self,
            layer_name,
            seg_id,
            points_red,
            points_blue,
            supervoxels_red=None,
            supervoxels_blue=None,
            focus=True,
        ):
            """Configures multicut points in the neuroglancer state. Note that points need to be in mesh units (e.g. nanometers), not voxels!

            Parameters
            ----------
            layer_name : str
                Segmentation layer name
            seg_id : np.uint64
                Segmentation id of the object in question
            points_red : np.array
                Nx3 array of locations in voxel space for side 1 of the cut.
            points_blue : np.array
                Mx3 array of locations in voxel space for side 2 of the cut.
            supervoxels_red : np.array or None, optional
                N-length array of supervoxel ids associated with locations in points_red or None. If None, supervoxels lookup occurs based on the mesh. By default None
            supervoxels_blue : np.array or None, optional
                M-length array of supervoxel ids associated with locations in points_blue or None. If None, supervoxels lookup occurs based on the mesh. By default None
            focus : bool, optional
                If True, makes the layer and graph tool focused. By default True
            """

            def _multicut_annotation(pt, oid, sv_id):
                if sv_id is None:
                    sv_id = oid
                return annotation.point_annotation(
                    pt, description=str(sv_id), linked_segmentation=[sv_id, oid]
                )

            if supervoxels_red is None:
                supervoxels_red = [None for x in points_red]
            if supervoxels_blue is None:
                supervoxels_blue = [None for x in points_blue]

            if self.state.voxel_size is None:
                voxel_size = [4, 4, 40]
            else:
                voxel_size = self.state.voxel_size

            annos_red = neuroglancer.annotationHolder()
            for pt, sv_id in zip(points_red, supervoxels_red):
                annos_red.annotations.append(
                    _multicut_annotation(pt * voxel_size, seg_id, sv_id)
                )

            annos_blue = neuroglancer.annotationHolder()
            for pt, sv_id in zip(points_blue, supervoxels_blue):
                annos_blue.annotations.append(
                    _multicut_annotation(pt * voxel_size, seg_id, sv_id)
                )

            self.add_selected_objects(layer_name, [seg_id])

            with self.txn() as s:
                l = s.layers[layer_name]
                l.tab = "graph"
                l.graphOperationMarker.append(annos_red)
                l.graphOperationMarker.append(annos_blue)

            if focus:
                self.set_selected_layer(layer_name)
                ctr_pt = vstack([points_red, points_blue]).mean(axis=0)
                self.set_view_options(position=ctr_pt, zoom_3d=100)

    return EasyViewerSub


def EasyViewer(compatibility_mode=False, **kwargs):
    """EasyViewer object that works with either the Seung Lab branch (default) or the
    Google main branch. Google branch support is highly experimental. It works for some basic features,
    but will not fail gracefully and many other features are unsupported. Use at your own risk!

    Parameters
    ----------
        compatibility_mode : bool, optional
            If False the EasyViewer works for Seung lab, if True for Google. Defaults to False.
    """
    EasyViewerCls = _EasyViewerFactory(compatibility_mode)
    EasyViewerCls.__name__ = "EasyViewer"
    return EasyViewerCls(**kwargs)


from neuroglancer.viewer_base import UnsynchronizedViewerBase
from neuroglancer.viewer_state import *
from neuroglancer.url_state import *


class Viewer(UnsynchronizedViewerBase):
    def __init__(self, **kwargs):
        super(UnsynchronizedViewerBase, self).__init__(**kwargs)

    def get_viewer_url(self):
        return "No server implemented"



import numpy as np
import pandas as pd
from ..easyviewer.base import SEGMENTATION_LAYER_TYPES


def layer_names(state):
    """Get all layer names in the state

    Parameters
    ----------
    state : dict
        Neuroglancer state as a JSON dict

    Returns
    -------
    names : list
        List of layer names
    """
    return [l["name"] for l in state["layers"]]


def image_layers(state):
    """Get all image layer names in the state

    Parameters
    ----------
    state : dict
        Neuroglancer state as a JSON dict

    Returns
    -------
    names : list
        List of layer names
    """
    return [l["name"] for l in state["layers"] if l["type"] == "image"]


def segmentation_layers(state):
    """Get all segmentation layer names in the state

    Parameters
    ----------
    state : dict
        Neuroglancer state as a JSON dict

    Returns
    -------
    names : list
        List of layer names
    """
    return [l["name"] for l in state["layers"] if l["type"] in SEGMENTATION_LAYER_TYPES]


def annotation_layers(state):
    """Get all annotation layer names in the state

    Parameters
    ----------
    state : dict
        Neuroglancer state as a JSON dict

    Returns
    -------
    names : list
        List of layer names
    """
    return [l["name"] for l in state["layers"] if l["type"] == "annotation"]


def tag_dictionary(state, layer_name):
    """Get the tag id to string dictionary for a layer

    Parameters
    ----------
    state : dict

    layer_name : [type]
        [description]

    Returns
    -------
    [type]
        [description]
    """
    l = get_layer(state, layer_name)
    taginfo = l.get("annotationTags", [])
    tags = {}
    for t in taginfo:
        tags[int(t["id"])] = t["label"]
    return tags


def get_layer(state, layer_name):
    """Gets the contents of the layer based on the layer name.

    Parameters
    ----------
    state : dict
        Neuroglancer state as a JSON dict
    layer_name : str
        Name of layer

    Returns
    -------
    layer : dict
        Layer data contents
    """
    layer_ind = layer_names(state).index(layer_name)
    return state["layers"][layer_ind]


def view_settings(state):
    """Get all data about the view state in neuroglancer: position,
    image zoom, orientation and zoom of the 3d view, and voxel size.

    Parameters
    ----------
    state : dict
        Neuroglancer state as JSON dict

    Returns
    -------
    view : dict
        Dictionary with keys: position, zoomFactor, perspectiveOrientation,
        perspectiveZoom, and voxelSize
    """
    view = {}
    view["position"] = state["navigation"]["pose"]["position"]["voxelCoordinates"]
    view["zoomFactor"] = state["navigation"].get("zoomFactor", None)
    view["perspectiveOrientation"] = state.get("perspectiveOrientation", None)
    view["perspectiveZoom"] = state.get("perspectiveZoom", None)
    view["voxelSize"] = state["navigation"]["pose"]["position"]["voxelSize"]
    return view


def get_selected_ids(state, layer=None):
    """Get a list of selected ids in a segmentation layer

    Parameters
    ----------
    state : dict
        State dict
    layer : str, optional
        Segmentation layer name, if needed. If None and only one segmentation layer
        is present, default to it. By default None

    Returns
    -------
    list
        List of root ids.
    """
    seg_layers = segmentation_layers(state)
    if len(seg_layers) == 1 and layer is None:
        layer = seg_layers[0]
    else:
        raise ValueError(
            "Segmentation layer must be specified since the state has more than one"
        )
    return [int(s) for s in get_layer(state, layer)["segments"]]


def _get_type_annotations(state, layer_name, type):
    l = get_layer(state, layer_name)
    annos = l.get("annotations", [])
    return [anno for anno in annos if anno["type"] == type]


def _get_point_annotations(state, layer_name):
    return _get_type_annotations(state, layer_name, "point")


def _get_sphere_annotations(state, layer_name):
    return _get_type_annotations(state, layer_name, "ellipsoid")


def _get_line_annotations(state, layer_name):
    return _get_type_annotations(state, layer_name, "line")


def _get_bbox_annotations(state, layer_name):
    return _get_type_annotations(state, layer_name, "line")


def _get_group_annotations(state, layer_name):
    return _get_type_annotations(state, layer_name, "collection")


def _extract_point_data(annos):
    return [[anno.get("point") for anno in annos]]


def _extract_sphere_data(annos):
    pt = [anno.get("center") for anno in annos]
    rad = [anno.get("radii") for anno in annos]
    return [pt, rad]


def _extract_line_data(annos):
    ptA = [anno.get("pointA") for anno in annos]
    ptB = [anno.get("pointB") for anno in annos]
    return [ptA, ptB]


def _extract_bbox_data(annos):
    ptA = [anno.get("pointA") for anno in annos]
    ptB = [anno.get("pointB") for anno in annos]
    return [ptA, ptB]


def _extract_group_data(annos):
    pt = [anno.get("source") for anno in annos]
    anno_id = [anno.get("id") for anno in annos]
    return [pt, anno_id]


_get_map = {
    "point": _get_point_annotations,
    "line": _get_line_annotations,
    "sphere": _get_sphere_annotations,
    "bbox": _get_bbox_annotations,
    "group": _get_group_annotations,
}

_extraction_map = {
    "point": _extract_point_data,
    "line": _extract_line_data,
    "sphere": _extract_sphere_data,
    "bbox": _extract_bbox_data,
    "group": _extract_group_data,
}


def _generic_annotations(
    state, layer_name, description, linked_segmentations, tags, group, anno_type
):
    annos = _get_map[anno_type](state, layer_name)
    out = _extraction_map[anno_type](annos)
    if description:
        desc = [anno.get("description", None) for anno in annos]
        out.append(desc)
    if linked_segmentations:
        linked_seg = [
            [np.uint64(x) for x in anno.get("segments", [])] for anno in annos
        ]
        out.append(linked_seg)
    if tags:
        tag_list = [anno.get("tagIds", []) for anno in annos]
        out.append(tag_list)
    if group:
        group_ids = [anno.get("parentId") for anno in annos]
        out.append(group_ids)
    if len(out) == 1:
        return out[0]
    else:
        return out


def point_annotations(
    state,
    layer_name,
    description=False,
    linked_segmentations=False,
    tags=False,
    group=False,
):
    """Get all point annotation points and other info from a layer.

    Parameters
    ----------
    state : dict
        Neuroglancer state as JSON dict
    layer_name : str
        Layer name
    description : bool, optional
        If True, also returns descriptions as well. By default False
    linked_segmentations : bool, optional
        If True, also returns list of linked segmentations, by default False
    tags : bool, optional
        If True, also returns list of tags, by default False

    Returns
    -------
    anno_points : list
        List of N 3-element points (as list)
    anno_descriptions : list
        List of N strings (or None), only returned if description=True.
    anno_linked_segmentations : list
        List of N lists of object ids. Only returned if linked_segmentations=True.
    anno_tags : list
        List of N lists of tag ids. Only returned if tags=True.
    anno_group : list
        List of group ids (as string) or None for annotations. Only returned if group=True
    """
    return _generic_annotations(
        state,
        layer_name,
        description=description,
        linked_segmentations=linked_segmentations,
        tags=tags,
        group=group,
        anno_type="point",
    )


def line_annotations(
    state,
    layer_name,
    description=False,
    linked_segmentations=False,
    tags=False,
    group=False,
):
    """Get all line annotation points and other info from a layer.

    Parameters
    ----------
    state : dict
        Neuroglancer state as JSON dict
    layer_name : str
        Layer name
    description : bool, optional
        If True, also returns descriptions as well. By default False
    linked_segmentations : bool, optional
        If True, also returns list of linked segmentations, by default False
    tags : bool, optional
        If True, also returns list of tags, by default False

    Returns
    -------
    anno_points_A : list
        List of N 3-element points (as list) of the first point in each line.
    anno_points_B : list
        List of N 3-element points (as list) of the second point in each line.
    anno_descriptions : list
        List of N strings (or None), only returned if description=True.
    anno_linked_segmentations : list
        List of N lists of object ids. Only returned if linked_segmentations=True.
    anno_tags : list
        List of N lists of tag ids. Only returned if tags=True.
    anno_group : list
        List of group ids (as string) or None for annotations. Only returned if group=True
    """
    return _generic_annotations(
        state, layer_name, description, linked_segmentations, tags, group, "line"
    )


def bbox_annotations(
    state,
    layer_name,
    description=False,
    linked_segmentations=False,
    tags=False,
    group=False,
):
    """Get all bounding box annotation points and other info from a layer.

    Parameters
    ----------
    state : dict
        Neuroglancer state as JSON dict
    layer_name : str
        Layer name
    description : bool, optional
        If True, also returns descriptions as well. By default False
    linked_segmentations : bool, optional
        If True, also returns list of linked segmentations, by default False
    tags : bool, optional
        If True, also returns list of tags, by default False

    Returns
    -------
    anno_points_A : list
        List of N 3-element points (as list) of the first point in each bbox.
    anno_points_B : list
        List of N 3-element points (as list) of the second point in each bbox.
    anno_descriptions : list
        List of N strings (or None), only returned if description=True.
    anno_linked_segmentations : list
        List of N lists of object ids. Only returned if linked_segmentations=True.
    anno_tags : list
        List of N lists of tag ids. Only returned if tags=True.
    anno_group : list
        List of group ids (as string) or None for annotations. Only returned if group=True
    """
    return _generic_annotations(
        state, layer_name, description, linked_segmentations, tags, group, "bbox"
    )


def sphere_annotations(
    state,
    layer_name,
    description=False,
    linked_segmentations=False,
    tags=False,
    group=False,
):
    """Get all sphere annotation points and other info from a layer.

    Parameters
    ----------
    state : dict
        Neuroglancer state as JSON dict
    layer_name : str
        Layer name
    description : bool, optional
        If True, also returns descriptions as well. By default False
    linked_segmentations : bool, optional
        If True, also returns list of linked segmentations, by default False
    tags : bool, optional
        If True, also returns list of tags, by default False

    Returns
    -------
    anno_points : list
        List of N 3-element center points (as list)
    radius_points : list
        List of N 3-element radii for each axis of the ellipsoid.
    anno_descriptions : list
        List of N strings (or None), only returned if description=True.
    anno_linked_segmentations : list
        List of N lists of object ids. Only returned if linked_segmentations=True.
    anno_tags : list
        List of N lists of tag ids. Only returned if tags=True.
    anno_group : list
        List of group ids (as string) or None for annotations. Only returned if group=True
    """
    return _generic_annotations(
        state, layer_name, description, linked_segmentations, tags, group, "sphere"
    )


def group_annotations(
    state, layer_name, description=False, linked_segmentations=False, tags=False
):
    """All group annotations and their associated points

    Parameters
    ----------
    state : dict
        Neuroglancer state as JSON dict
    layer_name : str
        Annotation layer name
    description : bool, optional
        If True, also returns descriptions as well. By default False
    linked_segmentations : bool, optional
        If True, also returns list of linked segmentations, by default False
    tags : bool, optional
        If True, also returns list of tags, by default False

    Returns
    -------
    group_points : list
        List of N 3-element points
    group_id: list
        List of N id strings for groups.
    anno_descriptions : list
        List of N strings (or None), only returned if description=True.
    anno_linked_segmentations : list
        List of N lists of object ids. Only returned if linked_segmentations=True.
    anno_tags : list
        List of N lists of tag ids. Only returned if tags=True.
    """
    return _generic_annotations(
        state,
        layer_name,
        description=description,
        linked_segmentations=linked_segmentations,
        tags=tags,
        group=False,
        anno_type="group",
    )


def extract_multicut(state, seg_layer=None):
    """Extract information entered into the multicut graph operation

    Parameters
    ----------
    state : dict
        Neuroglancer state
    seg_layer : str, optional
        Name of a segmentation layer or None. If None, the function will check how many segmentation
        layers there are and, if only one exits, choose it. If more than one segmentation layer is present,
        it errors. By default None

    Returns
    -------
    pts: np.array
        Nx3 array of points selected
    side: np.array
        N array with 'source' or 'sink', depending on which side the point is on.
    svids: np.array
        N array with selected supervoxel. If only points are selected (e.g. via clicking on the mesh),
        the value will be NaN.
    root_id: int
        Root id of the object to split
    """
    if seg_layer is None:
        seg_layers = segmentation_layers(state)
        if len(seg_layers) == 1:
            seg_layer = seg_layers[0]
        else:
            raise ValueError(
                "State has multiple segmentation layers. Please specify the layer to use."
            )
    l = get_layer(state, "seg")
    pts = []
    svids = []
    side = []
    root_id = None

    source_data = l["graphOperationMarker"][0]
    for anno in source_data["annotations"]:
        pts.append(anno["point"])
        side.append("source")
        if root_id is None:
            root_id = int(anno["segments"][1])
        svid = int(anno["segments"][0])
        if svid != root_id:
            svids.append(svid)
        else:
            # This means that no supervoxel was directly selected
            svids.append(np.nan)

    sink_data = l["graphOperationMarker"][1]
    for anno in sink_data["annotations"]:
        pts.append(anno["point"])
        side.append("sink")
        svid = int(anno["segments"][0])
        if root_id is None:
            root_id = int(anno["segments"][1])
        if svid != root_id:
            svids.append(svid)
        else:
            # This means that no supervoxel was directly selected
            svids.append(np.nan)

    return np.array(pts), np.array(side), np.array(svids), root_id


def annotation_dataframe(state):
    """Return all annotations across all annotation layers in the state.

    Parameters
    ----------
    state : dict
        Neuroglancer state dictionary

    Returns
    -------
    pd.DataFrame
        Dataframe with columns layer, anno_type, point, pointB, linked_segmentation, tags, anno_id, group_id, description.
    """

    lns = []
    points = []
    anno_types = []
    pointBs = []
    linked_segs = []
    tags = []
    group_ids = []
    descs = []

    for ln in annotation_layers(state):
        # Points
        p_pt, p_desc, p_seg, p_tag, p_grp = point_annotations(
            state,
            ln,
            description=True,
            linked_segmentations=True,
            tags=True,
            group=True,
        )
        n_p_pts = len(p_pt)
        p_type = ["point"] * n_p_pts
        p_ln = [ln] * n_p_pts
        p_ptB = [np.nan] * n_p_pts

        lns.append(p_ln)
        points.append(p_pt)
        anno_types.append(p_type)
        pointBs.append(p_ptB)
        linked_segs.append(p_seg)
        tags.append(p_tag)
        group_ids.append(p_grp)
        descs.append(p_desc)

        # Lines
        l_ptA, l_ptB, l_desc, l_seg, l_tag, l_grp = line_annotations(
            state,
            ln,
            description=True,
            linked_segmentations=True,
            tags=True,
            group=True,
        )
        n_l_pts = len(l_ptA)
        l_type = ["line"] * n_l_pts
        l_ln = [ln] * n_l_pts

        lns.append(l_ln)
        points.append(l_ptA)
        anno_types.append(l_type)
        pointBs.append(l_ptB)
        linked_segs.append(l_seg)
        tags.append(l_tag)
        group_ids.append(l_grp)
        descs.append(l_desc)

        # Spheres
        s_ptA, s_ptB, s_desc, s_seg, s_tag, s_grp = sphere_annotations(
            state,
            ln,
            description=True,
            linked_segmentations=True,
            tags=True,
            group=True,
        )
        n_s_pts = len(s_ptA)
        s_type = ["sphere"] * n_s_pts
        s_ln = [ln] * n_s_pts

        lns.append(s_ln)
        points.append(s_ptA)
        anno_types.append(s_type)
        pointBs.append(s_ptB)
        linked_segs.append(s_seg)
        tags.append(s_tag)
        group_ids.append(s_grp)
        descs.append(s_desc)

        # Bboxes
        b_ptA, b_ptB, b_desc, b_seg, b_tag, b_grp = bbox_annotations(
            state,
            ln,
            description=True,
            linked_segmentations=True,
            tags=True,
            group=True,
        )
        n_b_pts = len(b_ptA)
        b_type = ["bbox"] * n_b_pts
        b_ln = [ln] * n_b_pts

        lns.append(b_ln)
        points.append(b_ptA)
        anno_types.append(b_type)
        pointBs.append(b_ptB)
        linked_segs.append(b_seg)
        tags.append(b_tag)
        group_ids.append(b_grp)
        descs.append(b_desc)

    def _concat_list(d):
        d_out = []
        for x in d:
            for y in x:
                d_out.append(y)
        return d_out

    df = pd.DataFrame(
        {
            "layer": _concat_list(lns),
            "anno_type": _concat_list(anno_types),
            "point": _concat_list(points),
            "pointB": _concat_list(pointBs),
            "linked_segmentation": _concat_list(linked_segs),
            "tags": _concat_list(tags),
            "group_id": _concat_list(group_ids),
            "description": _concat_list(descs),
        }
    )
    return df



import numpy as np
from nglui import annotation
import pandas as pd
from collections.abc import Collection
from itertools import chain
from .utils import is_split_position, split_position_columns

def _multipoint_transform(row, pt_columns, squeeze_cols):
    """Reshape dataframe to accomodate multiple points in a single row"""
    pts = {pcol: np.atleast_2d(row[pcol]) for pcol in pt_columns}
    n_pts = pts[pt_columns[0]].shape[0]
    rows = [{} for _ in range(n_pts)]
    for col in row.index:
        if col in pt_columns:
            if col in squeeze_cols:
                for r, v in zip(rows, pts[col].squeeze().tolist()):
                    r[col] = v
            else:
                for r, v in zip(rows, pts[col].tolist()):
                    r[col] = v
        else:
            for r in rows:
                r[col] = row[col]
    return rows

def _multipoint_transform_split(df, multi_columns=[]):
    col_data = {}
    for col in df.columns:
        col_data[col] = []
    for _, row in df.iterrows():
        if len(multi_columns)>0:
            n_pts = len(row[multi_columns[0]])
        else:
            n_pts = 1
        for col in df.columns:
            if col in multi_columns:
                col_data[col].extend(row[col])
            else:
                col_data[col].extend(n_pts*[row[col]])
    return pd.DataFrame(col_data)


def _data_scaler(data_resolution, viewer_resolution):
    if viewer_resolution is None:
        return np.array([1, 1, 1]).reshape((1, 3))
    if data_resolution is None:
        data_resolution = viewer_resolution
    return (np.array(data_resolution) / np.array(viewer_resolution)).reshape((1, 3))


class SelectionMapper(object):
    """Class for configuring object selections based on root id

    Parameters
    ----------
    data_columns : str or list, optional
        Name (or list of names) of the data columns to get ids from. Default is None.
    fixed_ids : list, optional
        List of ids to select irrespective of data.
    fixed_id_colors : list, optional
        List of colors associated with the fixed ids list.
    color_column : str, optional
        Column name with color data per row
    mapping_set : str, optional
        If set, assumes data is passed as a dictionary and uses this string to as the key for the data to use.
        Note that using a mapping_set for one Mapper requires all Mappers use them. You cannot mix and match
        specificed mapping sets and ordered lists.
    """

    def __init__(
        self, data_columns=None, fixed_ids=None, fixed_id_colors=None, color_column=None, mapping_set=None,
    ):
        if isinstance(data_columns, str):
            data_columns = [data_columns]
        if fixed_id_colors is not None:
            fixed_id_colors = np.atleast_1d(fixed_id_colors).tolist()
        if fixed_ids is not None:
            fixed_ids = np.atleast_1d(fixed_ids).tolist()
        self._config = dict(
            data_columns=data_columns,
            fixed_ids=fixed_ids,
            fixed_id_colors=fixed_id_colors,
            color_column=color_column,
            mapping_set=mapping_set,
        )

    @property
    def data_columns(self):
        if self._config.get("data_columns", None) is None:
            return []
        else:
            return self._config.get("data_columns")

    @property
    def fixed_ids(self):
        if self._config.get("fixed_ids", None) is None:
            return np.atleast_1d(np.array([], dtype=np.uint64))
        else:
            return np.atleast_1d(
                np.array(self._config.get("fixed_ids", []), dtype=np.uint64)
            )

    @property
    def fixed_id_colors(self):
        if self._config.get("fixed_id_colors", None) is None:
            return []
        else:
            return list(self._config.get("fixed_id_colors", None))

    @property
    def color_column(self):
        return self._config.get("color_column", None)

    @property
    def mapping_set(self):
        return self._config.get("mapping_set", None)


    def selected_ids(self, data):
        """Uses the rules to generate a list of ids from a dataframe."""
        if self.mapping_set is not None:
            data = data.get(self.mapping_set)

        selected_ids = []

        if data is not None:
            for col in self.data_columns:
                selected_ids.append(data[col].values.astype(np.uint64))
        selected_ids.append(self.fixed_ids)
        return np.concatenate(selected_ids)

    def seg_colors(self, data):
        colors = {}
        if len(self.fixed_id_colors) == len(self.fixed_ids):
            for ii, oid in enumerate(self.fixed_ids):
                colors[oid] = self.fixed_id_colors[ii]

        if self.color_column is not None:
            clist = data[self.color_column].to_list()
            for col in self.data_columns:
                for ii, oid in enumerate(data[col]):
                    colors[oid] = clist[ii]

        return colors


class AnnotationMapperBase(object):
    def __init__(
        self,
        type,
        data_columns,
        description_column,
        linked_segmentation_column,
        tag_column,
        group_column,
        set_position,
        gather_linked_segmentations,
        share_linked_segmentations,
        multipoint,
        collapse_groups,
        split_positions=None,
        mapping_set=None,
    ):

        self._config = dict(
            type=type,
            data_columns=data_columns,
            array_data=False,
            description_column=description_column,
            linked_segmentation_column=linked_segmentation_column,
            tag_column=tag_column,
            group_column=group_column,
            set_position=set_position,
            gather_linked_segmentations=gather_linked_segmentations,
            share_linked_segmentations=share_linked_segmentations,
            multipoint=multipoint,
            collapse_groups=collapse_groups,
            split_positions=split_positions,
            mapping_set=mapping_set,
        )
        self._tag_map = None

    @property
    def type(self):
        return self._config.get("type", None)

    @property
    def data_columns(self):
        return self._config.get("data_columns", None)

    @property
    def description_column(self):
        return self._config.get("description_column", None)

    @property
    def linked_segmentation_column(self):
        return self._config.get("linked_segmentation_column", None)

    @property
    def tag_column(self):
        return self._config.get("tag_column", None)

    @property
    def group_column(self):
        return self._config.get("group_column", None)

    @property
    def gather_linked_segmentations(self):
        return self._config.get("gather_linked_segmentations", True)

    @property
    def share_linked_segmentations(self):
        return self._config.get("share_linked_segmentations", False)

    @property
    def set_position(self):
        return self._config.get("set_position", False)

    @property
    def multipoint(self):
        return self._config.get("multipoint", False)

    @property
    def collapse_groups(self):
        return self._config.get("collapse_groups", False)

    def multipoint_reshape(self, data, pt_columns, squeeze_cols=[]):
        if data is None or len(data) == 0:
            return data
        else:
            if not self.split_positions:
                rows = data.apply(
                    lambda x: _multipoint_transform(
                        x, pt_columns=pt_columns, squeeze_cols=squeeze_cols
                    ),
                    axis=1,
                ).tolist()
                return pd.DataFrame.from_records([r for r in chain.from_iterable(rows)])
            else:
                split_cols = []
                for pt_col in pt_columns:
                    split_cols.extend(split_position_columns(pt_col))
                return _multipoint_transform_split(data, split_cols)

    @property
    def split_positions(self):
        return self._config.get('split_positions')

    @property
    def mapping_set(self):
        return self._config.get('mapping_set', None)

    @property
    def tag_map(self):
        if self._tag_map is None:
            return {}
        else:
            return self._tag_map

    @tag_map.setter
    def tag_map(self, tag_list):
        if tag_list is None:
            self._tag_map = {}
        else:
            self._tag_map = {tag: ii + 1 for ii, tag in enumerate(tag_list)}

    @property
    def array_data(self):
        return self._config.get("array_data", False)

    @array_data.setter
    def array_data(self, new_array_data):
        self._config["array_data"] = new_array_data
        if new_array_data:
            self._config["data_columns"] = self._default_array_data_columns()
            self._config["description_column"] = None
            self._config["linked_segmentation_column"] = None
            self._config["tag_column"] = None

    def _default_array_data_columns(self):
        return []

    def _assign_tags(self, data):
        if self.tag_column is not None:
            anno_tags = []
            for row in data[self.tag_column]:
                if isinstance(row, Collection) and not isinstance(row, str):
                    add_annos = [self.tag_map.get(r, None) for r in row]
                else:
                    add_annos = [self.tag_map.get(row, None)]
                anno_tags.append(add_annos)
        else:
            anno_tags = [[None] for x in range(len(data))]
        return anno_tags


    def _parse_data(self, data, first=False):
        if self.mapping_set is not None:
            if not isinstance(data, dict):
                raise ValueError('If mapping sets are used, dataframes must be provided as values in a dictionary')
            data = data.get(self.mapping_set)
        if first and data is not None:
            if len(data) > 0:
                if isinstance(data, pd.DataFrame):
                    data = data.iloc[[0]]
                else:
                    data = np.array(data)[0].reshape(1,-1)
        return data

    def _process_columns(self, data, skip_columns=[]):
        if self.split_positions is not None:
            split_positions = self.split_positions
        else:
            is_split = np.bool([is_split(col, data) for col in self.data_columns])
            if np.all(~is_split):
                split_positions = False
            if np.any(is_split):
                split_positions = True
                # Check that the split applies the the correct columns, in case multiple are used
                for col, spl in zip(self.data_columns, is_split):
                    if spl:
                        skip_columns.append(col)
            
        if split_positions and not self.array_data:
            data = data.copy()
            for col in self.data_columns:
                if col in skip_columns:
                    continue
                split_cols = [f'{col}_{suf}' for suf in ["x", "y", "z"]]
                data[col] = np.vstack(data[split_cols].values).tolist()
            return data
        else:
            return data

    def _preprocess_data(self, data, skip_columns=[], squeeze_cols=[], first=False):
        data = self._parse_data(data, first=first)
        if data is None:
            return None

        if self.multipoint:
            data = self.multipoint_reshape(
                data, pt_columns=self.data_columns, squeeze_cols=squeeze_cols
            )

        data = self._process_columns(data, skip_columns=skip_columns)
        data = self._process_array_data(data)

        return data

    def _process_array_data(self, data):
        # Set per subclass
        return None

    def _render_data(self, data, data_resolution, viewer_resolution):
        # Set per subclass
        return None

    def _linked_segmentations(self, data):
        if self.linked_segmentation_column is not None:
            seg_array = data[self.linked_segmentation_column]
            linked_segs = [
                row if len(np.atleast_1d(row)) > 0 else None for row in seg_array.values
            ]
        else:
            linked_segs = [None for x in range(len(data))]
        return linked_segs

    def _descriptions(self, data):
        if self.description_column is not None:
            descriptions = data[self.description_column].values
        else:
            descriptions = [None for x in range(len(data))]
        return descriptions

    def _add_groups(self, data, annos):
        ngroups = data.groupby(self.group_column).ngroup().replace({-1: np.nan})
        vals, inverse = np.unique(ngroups, return_inverse=True)
        inv_inds = np.flatnonzero(~pd.isnull(vals))
        group_annos = []

        for ii in inv_inds:
            anno_to_group = [annos[jj] for jj in np.flatnonzero(inverse == ii)]
            group_annos.append(
                annotation.group_annotations(
                    anno_to_group,
                    return_all=False,
                    gather_linked_segmentations=self.gather_linked_segmentations,
                    share_linked_segmentations=self.share_linked_segmentations,
                    children_visible=not self.collapse_groups,
                )
            )
        annos.extend(group_annos)
        return annos


    def _get_position(self, data, data_resolution=None, viewer_resolution=None):
        # Parse data because get_position is called by the layer, not the render_annotation function.
        data = self._preprocess_data(data, skip_columns=self.data_columns[1:], first=True)
        if data is None:
            return None

        if len(data) > 0 and self.set_position is True:
            pt = np.atleast_2d(data[self.data_columns[0]].iloc[0])[0]
            if data_resolution is not None and viewer_resolution is not None:
                pt = pt * np.array(data_resolution) / np.array(viewer_resolution)
            return list(pt)
        else:
            return None


class PointMapper(AnnotationMapperBase):
    """Sets rules to map dataframes to point annotations

    Parameters
    ----------
    point_column : str, optional
        Column name with 3d position data
    array_data : bool, optional
        If True, the expected data is a Nx3 numpy array. This will only work for point positions,
        to use any other features you must build a dataframe.
    decription_column : str, optional
        Column name with string data for annotation descriptions
    linked_segmentation_column : str, optional
        Column name for root ids to link to annotations
    tag_column : str, optional
        Column name for categorical tag data. Tags must match those set in the
        annotation layer.
    group_column : str, optional
        Column name for grouping data. Data in this row should be numeric with possible NaNs.
        Rows with the same non-NaN value will be collected into a grouped annotation.
    set_position : bool, optional
        If set to True, moves the position to center on the first point in the
        data.
    multipoint: bool, optional
        If True, permits multiple points in a given row, sharing data in other columns. Default is False.
    collapse_groups: bool, optional
        If True, groups are toggled closed in the annotation view.
    mapping_set: str, optional
        If given, assumes data is passed as a dictionary and uses this string to as the key for the data to use.
    """

    def __init__(
        self,
        point_column=None,
        description_column=None,
        linked_segmentation_column=None,
        tag_column=None,
        group_column=None,
        gather_linked_segmentations=True,
        share_linked_segmentations=False,
        set_position=True,
        multipoint=False,
        collapse_groups=False,
        split_positions=False,
        mapping_set=None,
    ):
        super(PointMapper, self).__init__(
            type="point",
            data_columns=[point_column],
            description_column=description_column,
            linked_segmentation_column=linked_segmentation_column,
            tag_column=tag_column,
            group_column=group_column,
            gather_linked_segmentations=gather_linked_segmentations,
            share_linked_segmentations=share_linked_segmentations,
            set_position=set_position,
            multipoint=multipoint,
            collapse_groups=collapse_groups,
            split_positions=split_positions,
            mapping_set=mapping_set,
        )

    def _default_array_data_columns(self):
        return ["pt"]

    def _process_array_data(self, data):
        if self.array_data:
            data = pd.DataFrame(data={"pt": np.array(data).tolist()})
        return data

    def _render_data(self, data, data_resolution, viewer_resolution):
        "Transforms data to neuroglancer annotations"
        data = self._preprocess_data(data)
        if data is None:
            return []

        col = self.data_columns[0]
        relinds = ~pd.isnull(data[col])

        scaler = _data_scaler(data_resolution, viewer_resolution)
        pts = np.vstack(data[col][relinds]) * scaler
        descriptions = self._descriptions(data[relinds])

        linked_segs = self._linked_segmentations(data[relinds])
        tags = self._assign_tags(data)
        annos = [
            annotation.point_annotation(
                pt, description=d, linked_segmentation=ls, tag_ids=t
            )
            for pt, d, ls, t in zip(pts, descriptions, linked_segs, tags)
        ]
        if self.group_column is not None:
            annos = self._add_groups(data, annos)

        return annos


class LineMapper(AnnotationMapperBase):
    """Sets rules to map dataframes to line annotations

    Parameters
    ----------
    point_column_a : str
        Column name with 3d position data for the first point of the line.
        Must be set if array_data is False (the default)
    point_column_b : str
        Column name with 3d position data for the first point of the line.
        Must be set if array_data is False (the default)
    decription_column : str, optional
        Column name with string data for annotation descriptions
    linked_segmentation_column : str, optional
        Column name for root ids to link to annotations
    tag_column : str, optional
        Column name for categorical tag data. Tags must match those set in the annotation layer.
    group_column : str, optional
        Column name for grouping data. Data in this row should be numeric with possible NaNs.
        Rows with the same non-NaN value will be collected into a grouped annotation.
    set_position : bool, optional
        If set to True, moves the position to center on the first point in the data (using point_column_a).
    multipoint: bool, optional
        If True, permits multiple points in a given row, sharing data in other columns.
        Each point row must have the same number of points. Default is False.
    collapse_groups: bool, optional
        If True, groups are toggled closed in the annotation view.
    mapping_set: str, optional
        If set, assumes data is passed as a dictionary and uses this string to as the key for the data to use.
    """
    def __init__(
        self,
        point_column_a=None,
        point_column_b=None,
        description_column=None,
        linked_segmentation_column=None,
        tag_column=None,
        group_column=None,
        gather_linked_segmentations=True,
        share_linked_segmentations=False,
        set_position=True,
        multipoint=False,
        collapse_groups=False,
        split_positions=False,
        mapping_set=None,
    ):
        super(LineMapper, self).__init__(
            type="line",
            data_columns=[point_column_a, point_column_b],
            description_column=description_column,
            linked_segmentation_column=linked_segmentation_column,
            tag_column=tag_column,
            group_column=group_column,
            gather_linked_segmentations=gather_linked_segmentations,
            share_linked_segmentations=share_linked_segmentations,
            set_position=set_position,
            multipoint=multipoint,
            collapse_groups=collapse_groups,
            split_positions=split_positions,
            mapping_set=mapping_set,
        )

    def _default_array_data_columns(self):
        return ["pt_a", "pt_b"]

    def _process_array_data(self, data):
        if self.array_data:
            data = pd.DataFrame(
                data={"pt_a": data[0].tolist(), "pt_b": data[1].tolist()}
            )
        return data

    def _render_data(self, data, data_resolution, viewer_resolution):
        "Transforms data to neuroglancer annotations"
        data = self._preprocess_data(data)
        if data is None:
            return []

        colA, colB = self.data_columns
        relinds = np.logical_and(~pd.isnull(data[colA]), ~pd.isnull(data[colB]))

        scaler = _data_scaler(data_resolution, viewer_resolution)
        ptAs = np.vstack(data[colA][relinds]) * scaler
        ptBs = np.vstack(data[colB][relinds]) * scaler
        descriptions = self._descriptions(data[relinds])
        linked_segs = self._linked_segmentations(data[relinds])
        tags = self._assign_tags(data)
        annos = [
            annotation.line_annotation(
                ptA, ptB, description=d, linked_segmentation=ls, tag_ids=t
            )
            for ptA, ptB, d, ls, t in zip(ptAs, ptBs, descriptions, linked_segs, tags)
        ]

        if self.group_column is not None:
            annos = self._add_groups(data, annos)

        return annos


class SphereMapper(AnnotationMapperBase):
    """Sets rules to map dataframes to sphere annotations

    Parameters
    ----------
    center_column : str
        Column name with 3d position data for the center of the sphere
    radius_column : str
        Column name with a radius for the sphere (in nm)
    decription_column : str, optional
        Column name with string data for annotation descriptions
    linked_segmentation_column : str, optional
        Column name for root ids to link to annotations
    tag_column : str, optional
        Column name for categorical tag data. Tags must match those set in the annotation layer.
    group_column : str, optional
        Column name for grouping data. Data in this row should be numeric with possible NaNs.
        Rows with the same non-NaN value will be collected into a grouped annotation.
    set_position : bool, optional
        If set to True, moves the position to center on the first point in the data.
    multipoint: bool, optional
        If True, permits multiple points in a given row, sharing data in other columns.
        Each point row must have the same number of points. Default is False.
    collapse_groups: bool, optional
        If True, groups are toggled closed in the annotation view.
    mapping_set: str, optional
        If set, assumes data is passed as a dictionary and uses this string to as the key for the data to use.
    """

    def __init__(
        self,
        center_column=None,
        radius_column=None,
        description_column=None,
        linked_segmentation_column=None,
        tag_column=None,
        group_column=None,
        gather_linked_segmentations=True,
        share_linked_segmentations=False,
        z_multiplier=0.1,
        set_position=True,
        multipoint=False,
        collapse_groups=False,
        split_positions=False,
        mapping_set=None,
    ):
        super(SphereMapper, self).__init__(
            type="sphere",
            data_columns=[center_column, radius_column],
            description_column=description_column,
            linked_segmentation_column=linked_segmentation_column,
            tag_column=tag_column,
            group_column=group_column,
            gather_linked_segmentations=gather_linked_segmentations,
            share_linked_segmentations=share_linked_segmentations,
            set_position=set_position,
            multipoint=multipoint,
            collapse_groups=collapse_groups,
            split_positions=split_positions,
            mapping_set=mapping_set,
        )
        self._z_multiplier = z_multiplier

    def _default_array_data_columns(self):
        return ["ctr_pt", "rad"]

    def _process_array_data(self, data):
        if self.array_data:
            data = pd.DataFrame(data={"ctr_pt": data[0].tolist(), "rad": data[1]})
        return data

    def _render_data(self, data, data_resolution, viewer_resolution):
        "Transforms data to neuroglancer annotations"
        col_ctr, col_rad = self.data_columns

        data = self._preprocess_data(data, skip_columns=['rad'], squeeze_cols=[col_rad])
        if data is None:
            return []

        relinds = np.logical_and(~pd.isnull(data[col_ctr]), ~pd.isnull(data[col_rad]))

        scaler = _data_scaler(data_resolution, viewer_resolution)
        pts = np.vstack(data[col_ctr][relinds]) * scaler
        rs = data[col_rad][relinds].values

        if viewer_resolution:
            z_multiplier = viewer_resolution[1] / viewer_resolution[2]
        else:
            z_multiplier = self._z_multiplier

        descriptions = self._descriptions(data[relinds])
        linked_segs = self._linked_segmentations(data[relinds])
        tags = self._assign_tags(data)
        annos = [
            annotation.sphere_annotation(
                pt,
                r,
                description=d,
                linked_segmentation=ls,
                tag_ids=t,
                z_multiplier=z_multiplier,
            )
            for pt, r, d, ls, t in zip(pts, rs, descriptions, linked_segs, tags)
        ]

        if self.group_column is not None:
            annos = self._add_groups(data, annos)

        return annos


class BoundingBoxMapper(AnnotationMapperBase):
    """Sets rules to map dataframes to bounding box annotations

    Parameters
    ----------
    point_column_a : str
        Column name with 3d position data for the first point of the bounding box.
        Must be set if array_data is False (the default)
    point_column_b : str
        Column name with 3d position data for the second point of the bounding box.
        Must be set if array_data is False (the default)
    decription_column : str, optional
        Column name with string data for annotation descriptions
    linked_segmentation_column : str, optional
        Column name for root ids to link to annotations
    tag_column : str, optional
        Column name for categorical tag data. Tags must match those set in the annotation layer.
    group_column : str, optional
        Column name for grouping data. Data in this row should be numeric with possible NaNs.
        Rows with the same non-NaN value will be collected into a grouped annotation.
    set_position : bool, optional
        If set to True, moves the position to center on the first point in the data (using point_column_a).
    multipoint: bool, optional
        If True, permits multiple points in a given row, sharing data in other columns.
        Each point row must have the same number of points. Default is False.
    collapse_groups: bool, optional
        If True, groups are toggled closed in the annotation view.
    mapping_set: str, optional
        If set, assumes data is passed as a dictionary and uses this string to as the key for the data to use.
    """
    def __init__(
        self,
        point_column_a=None,
        point_column_b=None,
        description_column=None,
        linked_segmentation_column=None,
        tag_column=None,
        group_column=None,
        gather_linked_segmentations=True,
        share_linked_segmentations=False,
        set_position=True,
        multipoint=False,
        collapse_groups=False,
        split_positions=False,
        mapping_set=None,
    ):
        super(BoundingBoxMapper, self).__init__(
            type="axis_aligned_bounding_box",
            data_columns=[point_column_a, point_column_b],
            description_column=description_column,
            linked_segmentation_column=linked_segmentation_column,
            tag_column=tag_column,
            group_column=group_column,
            gather_linked_segmentations=gather_linked_segmentations,
            share_linked_segmentations=share_linked_segmentations,
            set_position=set_position,
            multipoint=multipoint,
            collapse_groups=collapse_groups,
            split_positions=split_positions,
            mapping_set=mapping_set,
        )

    def _default_array_data_columns(self):
        return ["pt_a", "pt_b"]

    def _process_array_data(self, data):
        if self.array_data:
            data = pd.DataFrame(
                data={"pt_a": data[0].tolist(), "pt_b": data[1].tolist()}
            )
        return data
    
    def _render_data(self, data, data_resolution, viewer_resolution):
        "Transforms data to neuroglancer annotations"
        data = self._preprocess_data(data)
        if data is None:
            return []

        colA, colB = self.data_columns
        relinds = np.logical_and(~pd.isnull(data[colA]), ~pd.isnull(data[colB]))

        scaler = _data_scaler(data_resolution, viewer_resolution)

        ptAs = np.vstack(data[colA][relinds]) * scaler
        ptBs = np.vstack(data[colB][relinds]) * scaler
        descriptions = self._descriptions(data[relinds])
        linked_segs = self._linked_segmentations(data[relinds])
        tags = self._assign_tags(data)
        annos = [
            annotation.bounding_box_annotation(
                ptA, ptB, description=d, linked_segmentation=ls, tag_ids=t
            )
            for ptA, ptB, d, ls, t in zip(ptAs, ptBs, descriptions, linked_segs, tags)
        ]

        if self.group_column is not None:
            annos = self._add_groups(data, annos)

        return annos


class SplitPointMapper(object):
    """Mapper to create split points in a segmentation layer.

    Parameters
    ----------
    id_column : str
        Column name for segment ids. The id column must contain the same id in all rows.
    point_column : str
        Name of the column containing points in space.
    team_column : str
        Name of the column describing the team for the points. The contents of the column should have two values, by default "red" and "blue".
    team_names : list, optional
        List of two values for the team names used in the team column. The first is mapped to red points, the second blue. Default is ["red", "blue"].
    supervoxel_column : str or None, optional
        Name of a column providing supervoxel ids. If None (default), the supervoxel must be looked up on the server.
    focus : bool, optional
        If True, sets the focus on the split tool and sets the position to the center of split points. Default is True.

    Returns
    -------
    SplitPointMapper instance to pass to a segmentation layer.
    """

    def __init__(
        self,
        id_column,
        point_column,
        team_column,
        team_names=["red", "blue"],
        supervoxel_column=None,
        focus=True,
        mapping_set=None,
    ):
        self.id_column = id_column
        self.point_column = point_column
        self.team_column = team_column
        self.team_names = team_names
        self.supervoxel_column = supervoxel_column
        self.focus = focus
        self.mapping_set = mapping_set

    def _render_data(self, df, data_resolution, viewer_resolution):
        if self.mapping_set is not None:
            df = df.get(self.mapping_set)

        if len(df) == 0:
            return None, np.atleast_2d([]), np.atleast_2d([]), []

        team_pts = []
        team_svs = []
        for team_name in self.team_names:
            team_df = df.query(f"{self.team_column} == @team_name")
            if len(team_df) > 0:
                team_pts.append(np.vstack(team_df[self.point_column].values))
                if self.supervoxel_column:
                    team_svs.append(team_df[self.supervoxel_column].values)
                else:
                    team_svs.append(None)
            else:
                team_pts.append(np.atleast_2d([]))

        seg_id = np.unique(df[self.id_column])
        if len(seg_id) > 1:
            raise ValueError("Multiple seg ids provided for SplitPointMapper")
        else:
            seg_id = seg_id[0]

        return seg_id, team_pts[0], team_pts[1], team_svs[0], team_svs[1]



from multiprocessing.sharedctypes import Value
from typing import Iterable
from .layers import (
    ImageLayerConfig,
    SegmentationLayerConfig,
    AnnotationLayerConfig,
    PointMapper,
    LineMapper,
)
from .statebuilder import ChainedStateBuilder, StateBuilder
from caveclient import CAVEclient
import pandas as pd
from IPython.display import HTML

DEFAULT_POSTSYN_COLOR = (0.25098039, 0.87843137, 0.81568627)  # CSS3 color turquise
DEFAULT_PRESYN_COLOR = (1.0, 0.38823529, 0.27843137)  # CSS3 color tomato

CONTRAST_CONFIG = {
    "minnie65_phase3_v1": {
        "contrast_controls": True,
        "black": 0.35,
        "white": 0.70,
    }
}
MAX_URL_LENGTH = 1_750_000
DEFAULT_NGL = "https://neuromancer-seung-import.appspot.com/"


def sort_dataframe_by_root_id(
    df, root_id_column, ascending=False, num_column="n_times", drop=False
):
    """Sort a dataframe so that rows belonging to the same root id are together, ordered by how many times the root id appears.

    Parameters
    ----------
    df : pd.DataFrame
        dataframe to sort
    root_id_column : str
        Column name to use for sorting root ids
    ascending : bool, optional
        Whether to sort ascending (lowest count to highest) or not, by default False
    num_column : str, optional
        Temporary column name to use for count information, by default 'n_times'
    drop : bool, optional
        If True, drop the additional column when returning.

    Returns
    -------
    pd.DataFrame
    """
    cols = df.columns[df.columns != root_id_column]
    if len(cols) == 0:
        df = df.copy()
        df[num_column] = df[root_id_column]
        use_column = num_column
    else:
        use_column = cols[0]

    df[num_column] = df.groupby(root_id_column).transform("count")[use_column]
    if drop:
        return df.sort_values(
            by=[num_column, root_id_column], ascending=ascending
        ).drop(columns=[num_column])
    else:
        return df.sort_values(by=[num_column, root_id_column], ascending=ascending)


def make_line_statebuilder(
    client: CAVEclient,
    point_column_a="pre_pt_position",
    point_column_b="post_pt_position",
    linked_seg_column="pt_root_id",
    description_column=None,
    tag_column=None,
    data_resolution=None,
    group_column=None,
    contrast=None,
    view_kws=None,
    point_layer_name="lines",
    color=None,
    split_positions=False,
):
    """make a state builder that puts points on a single column with a linked segmentaton id

    Args:
        client (CAVEclient): CAVEclient configured for the datastack desired
        point_column (str, optional): column in dataframe to pull points from. Defaults to "pt_position".
        linked_seg_column (str, optional): column to link to segmentation, None for no column. Defaults to "pt_root_id".
        group_columns (str, or list, optional): column(s) to group annotations by, None for no grouping (default=None)
        tag_column (str, optional): column to use for tags, None for no tags (default=None)
        description_column (str, optional): column to use for descriptions, None for no descriptions (default=None)
        contrast (list, optional):  list-like, optional
            Two elements specifying the black level and white level as
            floats between 0 and 1, by default None. If None, no contrast
            is set.
        view_kws (dict, optional): dict, optional
            dictionary of view keywords to configure neuroglancer view
        split_positions (bool, optional): whether the position column into x,y,z columns. Defaults to False.
    Returns:
        StateBuilder: a statebuilder to make points with linked segmentations
    """
    img_layer, seg_layer = from_client(client, contrast=contrast)
    line_mapper = LineMapper(
        point_column_a=point_column_a,
        point_column_b=point_column_b,
        linked_segmentation_column=linked_seg_column,
        tag_column=tag_column,
        description_column=description_column,
        group_column=group_column,
        split_positions=split_positions,
    )
    ann_layer = AnnotationLayerConfig(
        point_layer_name,
        mapping_rules=[line_mapper],
        linked_segmentation_layer=seg_layer.name,
        data_resolution=data_resolution,
        color=color,
    )
    if view_kws is None:
        view_kws = {}
    return StateBuilder(
        [img_layer, seg_layer, ann_layer],
        client=client,
        view_kws=view_kws,
    )


def make_point_statebuilder(
    client: CAVEclient,
    point_column="pt_position",
    linked_seg_column="pt_root_id",
    data_resolution=None,
    group_column=None,
    tag_column=None,
    description_column=None,
    contrast=None,
    view_kws=None,
    point_layer_name="pts",
    color=None,
    split_positions=False,
):
    """make a state builder that puts points on a single column with a linked segmentaton id

    Args:
        client (CAVEclient): CAVEclient configured for the datastack desired
        point_column (str, optional): column in dataframe to pull points from. Defaults to "pt_position".
        linked_seg_column (str, optional): column to link to segmentation, None for no column. Defaults to "pt_root_id".
        group_columns (str, or list, optional): column(s) to group annotations by, None for no grouping (default=None)
        tag_column (str, optional): column to use for tags, None for no tags (default=None)
        description_column (str, optional): column to use for descriptions, None for no descriptions (default=None)
        contrast (list, optional):  list-like, optional
            Two elements specifying the black level and white level as
            floats between 0 and 1, by default None. If None, no contrast
            is set.
        view_kws (dict, optional): dict, optional
            dictionary of view keywords to configure neuroglancer view
        split_positions (bool, optional): whether the position column into x,y,z columns. Defaults to False.
    Returns:
        StateBuilder: a statebuilder to make points with linked segmentations
    """
    img_layer, seg_layer = from_client(client, contrast=contrast)
    point_mapper = PointMapper(
        point_column=point_column,
        linked_segmentation_column=linked_seg_column,
        tag_column=tag_column,
        description_column=description_column,
        group_column=group_column,
        split_positions=split_positions,
    )
    ann_layer = AnnotationLayerConfig(
        point_layer_name,
        mapping_rules=[point_mapper],
        linked_segmentation_layer=seg_layer.name,
        data_resolution=data_resolution,
        color=color,
    )
    if view_kws is None:
        view_kws = {}
    return StateBuilder(
        [img_layer, seg_layer, ann_layer],
        client=client,
        view_kws=view_kws,
    )


def make_pre_post_statebuilder(
    client: CAVEclient,
    show_inputs: bool = False,
    show_outputs: bool = False,
    contrast: list = None,
    view_kws: dict = None,
    point_column="ctr_pt_position",
    pre_pt_root_id_col="pre_pt_root_id",
    post_pt_root_id_col="post_pt_root_id",
    dataframe_resolution_pre=None,
    dataframe_resolution_post=None,
    input_layer_name="syns_in",
    output_layer_name="syns_out",
    input_layer_color=DEFAULT_POSTSYN_COLOR,
    output_layer_color=DEFAULT_PRESYN_COLOR,
    split_positions=False,
):
    """Function to generate ChainedStateBuilder with optional pre and post synaptic
    annotation layers

    Args:
        client (CAVEclient): a CAVEclient configured for datastack to visualize
        root_id (int): a rootID to build around
        show_inputs (bool, optional): whether to show input synapses. Defaults to False.
        show_outputs (bool, optional): whether to show output synapses.. Defaults to False.
        contrast (list, optional):  list-like, optional
            Two elements specifying the black level and white level as
            floats between 0 and 1, by default None. If None, no contrast
            is set.
        view_kws (dict, optional): view_kws to configure statebuilder, see nglui.StateBuilder . Defaults to None.
            keys are:
                show_slices: Boolean
                    sets if slices are shown in the 3d view. Defaults to False.
                layout: str
                    `xy-3d`/`xz-3d`/`yz-3d` (sections plus 3d pane), `xy`/`yz`/`xz`/`3d` (only one pane), or `4panel` (all panes). Default is `xy-3d`.
                show_axis_lines: Boolean
                    determines if the axis lines are shown in the middle of each view.
                show_scale_bar: Boolean
                    toggles showing the scale bar.
                orthographic : Boolean
                    toggles orthographic view in the 3d pane.
                position* : 3-element vector
                    determines the centered location.
                zoom_image : float
                    Zoom level for the imagery in units of nm per voxel. Defaults to 8.
                zoom_3d : float
                    Zoom level for the 3d pane. Defaults to 2000. Smaller numbers are more zoomed in.
        point_column (str, optional): column to pull points for synapses from. Defaults to "ctr_pt_position".
        pre_pt_root_id_col (str, optional): column to pull pre synaptic ids for synapses from. Defaults to "pre_pt_root_id".
        post_pt_root_id_col (str, optional): column to pull post synaptic ids for synapses from. Defaults to "post_pt_root_id".
        input_layer_name (str, optional): name of layer for inputs. Defaults to "syns_in".
        output_layer_name (str, optional): name of layer for outputs. Defaults to "syns_out".
        split_positions (bool, optional): whether the position column is split into x,y,z columns. Defaults to False.
    Returns:
        ChainedStateBuilder: An instance of a ChainedStateBuilder configured to accept
        a list  starting with None followed by optionally synapse input dataframe
        followed by optionally synapse output dataframe.
    """

    img_layer, seg_layer = from_client(client, contrast=contrast)
    seg_layer.add_selection_map(selected_ids_column="root_id")

    if view_kws is None:
        view_kws = {}
    sb1 = StateBuilder(layers=[img_layer, seg_layer], client=client, view_kws=view_kws)

    state_builders = [sb1]
    if show_inputs:
        # First state builder
        input_point_mapper = PointMapper(
            point_column=point_column,
            linked_segmentation_column=pre_pt_root_id_col,
            split_positions=split_positions,
        )
        inputs_lay = AnnotationLayerConfig(
            input_layer_name,
            mapping_rules=[input_point_mapper],
            linked_segmentation_layer=seg_layer.name,
            data_resolution=dataframe_resolution_post,
            color=input_layer_color,
        )
        sb_in = StateBuilder([inputs_lay], client=client)
        state_builders.append(sb_in)
    if show_outputs:
        output_point_mapper = PointMapper(
            point_column=point_column,
            linked_segmentation_column=post_pt_root_id_col,
            split_positions=split_positions,
        )
        outputs_lay = AnnotationLayerConfig(
            output_layer_name,
            mapping_rules=[output_point_mapper],
            linked_segmentation_layer=seg_layer.name,
            data_resolution=dataframe_resolution_pre,
            color=output_layer_color,
        )
        sb_out = StateBuilder([outputs_lay], client=client)
        state_builders.append(sb_out)
    return ChainedStateBuilder(state_builders)


def make_state_url(df, sb, client, ngl_url=None):
    state = sb.render_state(df, return_as="dict")
    state_id = client.state.upload_state_json(state)
    if ngl_url is None:
        ngl_url = client.info.viewer_site()
        if ngl_url is None:
            ngl_url = DEFAULT_NGL
    url = client.state.build_neuroglancer_url(state_id, ngl_url=ngl_url)
    return url


def make_url_robust(
    df: pd.DataFrame,
    sb: StateBuilder,
    client: CAVEclient,
    shorten: str = "if_long",
    ngl_url: str = None,
    max_url_length = MAX_URL_LENGTH,
):
    """Generate a url from a neuroglancer state. If too long, return through state server

    Args:
        df (pandas.DataFrame): dataframe to pass through statebuilder
        sb (nglui.statebuilder.StateBuilder): statebuilder to generate link with
        client (caveclient.CAVEclient): client to interact with state server with and get defaults from
        shorten (str, optional): How to shorten link. one of 'if_long', 'always', 'never'.
            'if_long' will use the state server to shorten links longer than nglui.statebuilder.MAX_URL_LENGTH
            (set to 1,750,000).
            'always' will always use the state server to shorten the url
            'never' will always return the full url.  Defaults to "if_long".
        ngl_url (str, optional): neuroglancer deployment to make url with.
            Defaults to None, which will use the default in the passed sb StateBuilder

    Returns:
        str: a url containing the state created by the statebuilder.
    """

    if shorten == "if_long":
        url = sb.render_state(df, return_as="url", url_prefix=ngl_url)
        if len(url) > max_url_length:
            url = make_state_url(df, sb, client, ngl_url=ngl_url)
    elif shorten == "always":
        url = make_state_url(df, sb, client)
    elif shorten == "never":
        url = sb.render_state(df, return_as="url", url_prefix=ngl_url)
    else:
        raise (ValueError('shorten should be one of ["if_long", "always", "never"]'))
    return url


def package_state(
    df: pd.DataFrame,
    sb: StateBuilder,
    client: CAVEclient,
    shorten: str = "if_long",
    return_as: str = "url",
    ngl_url: str = None,
    link_text: str = "Neuroglancer Link",
):
    """a function to automate creating a state from a statebuilder and
    a dataframe, return it in the desired format, shortening if desired.

    Args:
        df (pd.DataFrame): dataframe to pass to the sb:StateBuilder
        sb (StateBuilder): StateBuilder to generate links with
        client (CAVEclient): caveclient to get default ngl_url and iteract with state viewer
        shorten (str): one of ["if_long", "always", "never"]
            'if_long' will use the state server to shorten links longer than nglui.statebuilder.MAX_URL_LENGTH
            (set to 1,750,000).
            'always' will always use the state server to shorten the url
            'never' will always return the full url.  Defaults to "if_long".
        return_as (str): one of ['html', 'url', 'json']
            'html' will return an ipython clickable link
            'url' will return a string with the url
            'json' will return the state as a dictionary
        ngl_url (str): neuroglancer url to use, if None will use client.info.viewer_site()
        link_text (str): if returning as html, what text to show for the link.

    Returns:
       HTML, str or dict : state in format specified by return_as
    """
    if ngl_url is None:
        ngl_url = client.info.viewer_site()
        if ngl_url is None:
            ngl_url = DEFAULT_NGL

    if (return_as == "html") or (return_as == "url"):
        url = make_url_robust(df, sb, client, shorten=shorten, ngl_url=ngl_url)
        if return_as == "html":
            return HTML(f'<a href="{url}">{link_text}</a>')
        else:
            return url
    elif return_as == "json":
        return sb.render_state(df, return_as=return_as)
    else:
        raise (
            ValueError(
                f'return_as={return_as} not a valid choice, choose one of "html", "url", or "json")'
            )
        )


def make_synapse_neuroglancer_link(
    synapse_df,
    client,
    return_as="html",
    shorten="always",
    contrast=None,
    point_column="ctr_pt_position",
    dataframe_resolution=None,
    group_connections=True,
    link_pre_and_post=True,
    ngl_url=None,
    view_kws=None,
    pre_post_columns=None,
    neuroglancer_link_text="Neuroglancer Link",
    color=None,
    split_positions=False,
):
    """_summary_

    Args:
        synapse_df (pandas.DataFrame): dataframe where each row is a synapse
        client (caveclient.CAVEclient): a caveclient
        return_as (str, optional): how to return url.
            'html' as a ipython html element
            'url' as a url string
            'json' as dictionary state.
            Defaults to "html".
        shorten (str, optional): whether to shorten link, 'always' will always shorten,
             'if_long' will shorten it if it's beyond a size MAX_URL_LENGTH.
             'never' will never shorten. Defaults to "always".
        contrast (list, optional): Two elements specifying the black level and white level as
            floats between 0 and 1, by default None. If None, no contrast is set.
        point_column (str, optional): column in dataframe to take synapse position from. Defaults to "ctr_pt_position".
        dataframe_resolution (list, optional): list of length 3 specifying resolution units of position column.
            If None, will attempt to get resolution from dataframe metadata. If no metadata exists,
            will assume it's in client.info.viewer_resolution(). Defaults to None.
        group_connections (bool, optional): whether to group synapses in the same connection (between the same neurons).
            Defaults to True.
        link_pre_and_post (bool, optional): whether to make link the synapse annotations
            to the pre and post synaptic partners. Defaults to True.
        ngl_url (_type_, optional): what neuroglancer to use. Defaults to None.
            If None will default to client.info.viewer_site().
            If that is None will default to DEFAULT_NGL.
        view_kws (dict, optional): viewer dictionary to control neuroglancer viewer. Defaults to None.
            keys are:
                show_slices: Boolean
                    sets if slices are shown in the 3d view. Defaults to False.
                layout: str
                    `xy-3d`/`xz-3d`/`yz-3d` (sections plus 3d pane), `xy`/`yz`/`xz`/`3d` (only one pane), or `4panel` (all panes). Default is `xy-3d`.
                show_axis_lines: Boolean
                    determines if the axis lines are shown in the middle of each view.
                show_scale_bar: Boolean
                    toggles showing the scale bar.
                orthographic : Boolean
                    toggles orthographic view in the 3d pane.
                position* : 3-element vector
                    determines the centered location.
                zoom_image : float
                    Zoom level for the imagery in units of nm per voxel. Defaults to 8.
                zoom_3d : float
                    Zoom level for the 3d pane. Defaults to 2000. Smaller numbers are more zoomed in.
        pre_post_columns (list, optional): [pre,post] column names for pre and post synaptic root_ids. Defaults to None.
            If None will assume ['pre_pt_root_id', 'post_pt_root_id']
        neuroglancer_link_text (str, optional): Text to use in returning html link. Defaults to "Neuroglancer Link".
        color (list(float) or str, optional): color of synapse points as rgb list [0,1],
            or hex string, or common name (see webcolors documentation)
        split_positions (bool, optional): whether the position column are splits into x,y,z columns.

    Raises:
        ValueError: If the point_column is not in the dataframe

    Returns:
        Ipython.HTML, str, or json: a representation of the neuroglancer state.Type depends on return_as
    """
    if point_column not in synapse_df.columns:
        raise ValueError(f"point_column={point_column} not in dataframe")
    if pre_post_columns is None:
        pre_post_columns = ["pre_pt_root_id", "post_pt_root_id"]
    if dataframe_resolution is None:
        dataframe_resolution = synapse_df.attrs.get("dataframe_resolution", None)

    if group_connections:
        group_column = pre_post_columns
    else:
        group_column = None
    if link_pre_and_post:
        linked_columns = pre_post_columns
    else:
        linked_columns = None

    sb = make_point_statebuilder(
        client,
        point_column=point_column,
        linked_seg_column=linked_columns,
        group_column=group_column,
        data_resolution=dataframe_resolution,
        contrast=contrast,
        view_kws=view_kws,
        point_layer_name="synapses",
        color=color,
        split_positions=split_positions,
    )
    return package_state(
        synapse_df, sb, client, shorten, return_as, ngl_url, neuroglancer_link_text
    )


def make_neuron_neuroglancer_link(
    client,
    root_ids,
    return_as="html",
    shorten="always",
    show_inputs=False,
    show_outputs=False,
    sort_inputs=True,
    sort_outputs=True,
    sort_ascending=False,
    input_color=DEFAULT_POSTSYN_COLOR,
    output_color=DEFAULT_PRESYN_COLOR,
    contrast=None,
    timestamp=None,
    view_kws=None,
    point_column="ctr_pt_position",
    pre_pt_root_id_col="pre_pt_root_id",
    post_pt_root_id_col="post_pt_root_id",
    input_layer_name="syns_in",
    output_layer_name="syns_out",
    ngl_url=None,
    link_text="Neuroglancer Link",
):
    """function to create a neuroglancer link view of a neuron, optionally including inputs and outputs

    Args:
        client (_type_): a CAVEclient configured for datastack to visualize
        root_ids (Iterable[int]): root_ids to build around
        return_as (str, optional): one of 'html', 'json', 'url'. (default 'html')
        shorten (str, optional): if 'always' make a state link always
                             'if_long' make a state link if the json is too long (default)
                             'never' don't shorten link
        show_inputs (bool, optional): whether to include input synapses. Defaults to False.
        show_outputs (bool, optional): whether to include output synapses. Defaults to False.
        sort_inputs (bool, optional): whether to sort inputs by presynaptic root id, ordered by synapse count.
            Defaults to True.
        sort_outputs (bool, optional): whether to sort inputs by presynaptic root id, ordered by postsynaptic synapse count.
            Defaults to True.
        sort_ascending (bool, optional): If sorting, whether to sort ascending (lowest synapse count to highest).
            Defaults to False.
        input_color (list(float) or str, optional): color of input points as rgb list [0,1],
            or hex string, or common name (see webcolors documentation)
        output_color (list(float) or str, optional): color of output points as rgb list [0,1],
            or hex string, or common name (see webcolors documentation)
        contrast (list, optional): Two elements specifying the black level and white level as
            floats between 0 and 1, by default None. If None, no contrast is set.
        timestamp (datetime.datetime, optional): timestamp to do query. Defaults to None, will use materialized version.
        view_kws (dict, optional): view_kws to configure statebuilder, see nglui.StateBuilder.
            Defaults to None.
            keys are:
                show_slices: Boolean
                    sets if slices are shown in the 3d view. Defaults to False.
                layout: str
                    `xy-3d`/`xz-3d`/`yz-3d` (sections plus 3d pane), `xy`/`yz`/`xz`/`3d` (only one pane), or `4panel` (all panes). Default is `xy-3d`.
                show_axis_lines: Boolean
                    determines if the axis lines are shown in the middle of each view.
                show_scale_bar: Boolean
                    toggles showing the scale bar.
                orthographic : Boolean
                    toggles orthographic view in the 3d pane.
                position* : 3-element vector
                    determines the centered location.
                zoom_image : float
                    Zoom level for the imagery in units of nm per voxel. Defaults to 8.
                zoom_3d : float
                    Zoom level for the 3d pane. Defaults to 2000. Smaller numbers are more zoomed in.
        point_column (str, optional): column to pull points for synapses from. Defaults to "ctr_pt_position".
        pre_pt_root_id_col (str, optional): column to pull pre synaptic ids for synapses from.
            Defaults to "pre_pt_root_id".
        post_pt_root_id_col (str, optional): column to pull post synaptic ids for synapses from.
            Defaults to "post_pt_root_id".
        input_layer_name (str, optional): name of layer for inputs. Defaults to "syns_in".
        output_layer_name (str, optional): name of layer for outputs. Defaults to "syns_out".
        ngl_url (str, optional): url to use for neuroglancer.
            Defaults to None (will use default viewer set in datastack)
        link_text (str, optional): text to use for html return.
            Defaults to 'Neuroglancer Link'
    Raises:
        ValueError: If the point column is not present in the synapse table

    Returns:
        Ipython.HTML, str, or json: a representation of the neuroglancer state.Type depends on return_as
    """
    if not isinstance(root_ids, Iterable):
        root_ids = [root_ids]
    df1 = pd.DataFrame({"root_id": root_ids})
    dataframes = [df1]
    data_resolution_pre = None
    data_resolution_post = None
    if show_inputs:
        syn_in_df = client.materialize.synapse_query(
            post_ids=root_ids,
            timestamp=timestamp,
            desired_resolution=client.info.viewer_resolution(),
            split_positions=True,
        )
        data_resolution_pre = syn_in_df.attrs["dataframe_resolution"]
        if sort_inputs:
            syn_in_df = sort_dataframe_by_root_id(
                syn_in_df, pre_pt_root_id_col, ascending=sort_ascending, drop=True
            )
        # if point_column not in syn_in_df.columns:
            # raise ValueError("column pt_column={pt_column} not in synapse table")
        dataframes.append(syn_in_df)
    if show_outputs:
        syn_out_df = client.materialize.synapse_query(
            pre_ids=root_ids,
            timestamp=timestamp,
            desired_resolution=client.info.viewer_resolution(),
            split_positions=True,
        )
        data_resolution_post = syn_out_df.attrs["dataframe_resolution"]
        if sort_outputs:
            syn_out_df = sort_dataframe_by_root_id(
                syn_out_df, post_pt_root_id_col, ascending=sort_ascending, drop=True
            )
        dataframes.append(syn_out_df)
    sb = make_pre_post_statebuilder(
        client,
        show_inputs=show_inputs,
        show_outputs=show_outputs,
        contrast=contrast,
        point_column=point_column,
        view_kws=view_kws,
        pre_pt_root_id_col=pre_pt_root_id_col,
        post_pt_root_id_col=post_pt_root_id_col,
        input_layer_name=input_layer_name,
        output_layer_name=output_layer_name,
        input_layer_color=input_color,
        output_layer_color=output_color,
        dataframe_resolution_pre=data_resolution_pre,
        dataframe_resolution_post=data_resolution_post,
        split_positions=True,
    )
    return package_state(dataframes, sb, client, shorten, return_as, ngl_url, link_text)


def from_client(client, image_name=None, segmentation_name=None, contrast=None):
    """Generate basic image and segmentation layers from a FrameworkClient

    Parameters
    ----------
    client : annotationframeworkclient.FrameworkClient
        A FrameworkClient with a specified datastack
    image_name : str, optional
        Name for the image layer, by default None.
    segmentation_name : str, optional
        Name for the segmentation layer, by default None
    contrast : list-like, optional
        Two elements specifying the black level and white level as
        floats between 0 and 1, by default None. If None, no contrast
        is set.

    Returns
    -------
    image_layer : ImageLayerConfig
        Image layer with default values from the client
    seg_layer : ImageLayerConfig
        Segmentation layer with default values from the client
    """
    if contrast is None:
        config = CONTRAST_CONFIG.get(
            client.datastack_name, {"contrast_controls": True, "black": 0, "white": 1}
        )
    else:
        config = {"contrast_controls": True, "black": contrast[0], "white": contrast[1]}
    img_layer = ImageLayerConfig(client.info.image_source(), name=image_name, **config)
    seg_layer = SegmentationLayerConfig(
        client.info.segmentation_source(), name=segmentation_name
    )
    return img_layer, seg_layer



from .. import nglite as neuroglancer
import numpy as np
import pandas as pd
import webcolors
import re
import numbers

SegmentationLayer = neuroglancer.viewer_state.SegmentationLayer

default_neuroglancer_base = "https://neuromancer-seung-import.appspot.com/"


def omit_nones(seg_list):
    if seg_list is None or np.all(pd.isna(seg_list)):
        return []
    seg_list = np.atleast_1d(seg_list)
    seg_list = list(filter(lambda x: x is not None, seg_list))
    if len(seg_list) == 0:
        return []
    else:
        return seg_list


def parse_color(clr):
    if clr is None:
        return None

    if isinstance(clr, numbers.Number):
        clr = (clr, clr, clr)

    if isinstance(clr, str):
        hex_match = "\#[0123456789abcdef]{6}"
        if re.match(hex_match, clr.lower()):
            return clr
        else:
            return webcolors.name_to_hex(clr)
    else:
        return webcolors.rgb_to_hex([int(255 * x) for x in clr])



from .helpers import from_client, make_neuron_neuroglancer_link
from .statebuilder import StateBuilder, ChainedStateBuilder
from .layers import SegmentationLayerConfig, AnnotationLayerConfig, ImageLayerConfig
from .mappers import (
    SelectionMapper,
    PointMapper,
    LineMapper,
    SphereMapper,
    BoundingBoxMapper,
    SplitPointMapper,
)
